@@ -97,6 +97,8 @@ public:
 
     void setEncodingPreference(const PipeWireBaseEncodedStream::EncodingPreference &encodingPreference);
 
+    void handleEncodedFramesChanged();
+
     const uint m_nodeId;
     QScopedPointer<PipeWireSourceStream> m_stream;
     QString m_error;
@@ -126,8 +128,10 @@ public:
     std::atomic_bool m_passthroughRunning = false;
     std::atomic_bool m_outputRunning = false;
 
-    std::condition_variable m_frameReceivedCondition;
-    std::mutex m_frameReceivedMutex;
+    std::condition_variable m_passthroughCondition;
+    std::mutex m_passthroughMutex;
+    std::condition_variable m_outputCondition;
+    std::mutex m_outputMutex;
 
     std::atomic_bool m_deactivated = false;
 
